/*****************************************************************************/
/*
                                 DynDNSupd.c

Light-weight, set-and-forget DynDNS DDNS update client for OpenVMS.

  http://www.dyndns.com/services/dns/dyndns/

  "The free Dynamic DNS service allows you to alias a dynamic IP address
   to a static hostname in any of the many domains we offer, allowing your
   computer to be more easily accessed from various locations on the Internet.
   We provide this service, for up to five (5) hostnames, free to the Internet
   community.

   The Dynamic DNS service is ideal for a home website, file server, or just
   to keep a pointer back to your home PC so you can access those important
   documents while you're at work. Using one of the available third-party
   update clients [this!] you can keep your hostname always pointing to your
   IP address, no matter how often your ISP changes it. No more fumbling to
   find that piece of paper where you wrote do wn your IP address, or
   e-mailing all your friends every time it changes. Just tell them to visit
   yourname.dyndns.org instead!"
 
Based on the specifications available from

  http://www.dyndns.com/developers/specs/

Updates are (only) performed using SSL and so DynDNS credentials are secure
during the transaction.  This does mean the HP (Open) SSL product must be
installed and started on the intended system.

The system current IP address is established by using the DynDNS
checkip.dyndns.org Web service.  Only when this shows an IP address change the
members.dyndns.org server is used to attempt/perform an update.

Most of the run-time parameters that can be defaulted, have been.  For example;
the members.dyndns.org update host, the check period of thirty minutes, etc. 
The parameters that are site specific, such as the members DynDNS dynamic host
name, the member authorization credentials, etc., need to be supplied using the
appropriate qualifier.  These parameters can be provided at the command-line or
using logical name values.  This latter mechanism is the intended method of
operation.

DynDNSupd is designed to be run as a detached process and obtaining it's
parameters from logical name values.  It also provides status information back
via the values of logical names (see below).  In addition it modifies it's
process name to provide a very granular indication of the update status.

  DynDNSupd:OK      an update has been successfully made
  DynDNSupd:FAIL    an attempted update has failed
  DynDNSupd:NOCHG   no update was necessary (no detected change in IP address)
  DynDNSupd:ERROR   some other error has occured during the IP-check/update

Specifics of the failure may be found in the (recommended) log file.


STARTUP
-------
Startup for the utility should be placed into the system startup procedure. 
The following provides a minimal but functional configuration.

  $ DEFINE /SYSTEM /EXECUTIVE DYNDNSUPD_AUTH "username:password"
  $ DEFINE /SYSTEM /EXECUTIVE DYNDNSUPD_PARAM "/HOST=myhost.dyndns.org /OPCOM"
  $ DEFINE /SYSTEM /EXECUTIVE DYNDNSUPD_LOG dev:[dir]DYNDNSUPD.LOG 
  $ RUN /DETACHED dev:[dir]DYNDNSUPD.EXE

The DYNDNSUPD_AUTH logical name is automatically deleted once the utility has
read the value.  If the detached process is deleted this logical name and value
must be redfined before restarting the utility.  The DYNDNSUPD_PARAM must
remain for the life of the utility executing in the detached process.  The
value is read before each IP address check (and possible update).  This does
imply that the contents may be modified between during the check/update period.

The log file of the detached process contains all significant events and errors
during utility processing.  It is flushed just before hibernating after each
check/update and so should be up-to-date at most times.


COMMAND-LINE PARAMETERS
-----------------------
/AUTH=<username:pwd>      DynDNS credentials to authenticate the update
                            (in normal use DYNDNSUPD_AUTH logical)
/BACKMX=<YES|NO|NOCHG>    DynDNS backup mail exchanger parameter
/DBUG                     low-level debug statements
/FORCE=<days>             unconditional update used to ensure the ownership
                            of the host(s) is maintained
/HOSTNAME=<host[,host]>   DynDNS dynamic host(s) to have their IP updated
/MX=<host|NOCHG>          DynDNS mail exchanger parameter
/NOLOG                    override the presence of a DYNDNSUPD_LOG logical
/OPCOM[=<integer>]        decimal number of OPCOM target, defaults to CENTRAL
/PERIOD=<mins>|<hh:mm>    check/update period in minutes or hours:minutes
/NOW                      force the update (one-shot)
/WILDCARD=<ON|OFF|NOCHG>  DynDNS wildcard parameter
/WATCH                    high-level debug statements


LOGICAL NAMES
-------------
All logical names must be and are created /SYSTEM /EXECUTIVE.

Control functionality:

DYNDNSUPD_AUTH          DynDNS username and password
DYNDNSUPD_HOST_CHECK    non-default host:port for IP address check
DYNDNSUPD_HOST_UPDATE   non-default host:port for DynDNS update
DYNDNSUPD_LOG           location and name of log file (if this ends in ";0"
                          any existing log file is appended to)
DYNDNSUPD_NOW           if this exists the next update is forced and
                          at the same time the logical name deleted
DYNDNSUPD_PARAM         command-line startup parameters

Debugging functionality:

DYNDNSUPD$DBUG          same as /DBUG
DYNDNSUPD$WATCH         same as /WATCH

Created in LNM$SYSTEM by DYNDNSUPD:

DYNDNSUPD_IP            the latest IP address resolved for the host
DYNDNSUPD_PID           the PID of the detached process as hex ASCII
DYNDNSUPD_RETURN        information about return code from latest update
DYNDNSUPD_TIME_LAST     time string of latest check of IP address
DYNDNSUPD_TIME_NEXT     time string of next check of IP address
DYNDNSUPD_TIME_UPDATE   time string of latest update of IP address


BUILD DETAILS
-------------
Compile then link:
  $ @BUILD_DYNDNSUPD
To just link:
  $ @BUILD_DYNDNSUPD LINK


COPYRIGHT
---------
Copyright (C) 2006 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY (update SOFTWAREVN as well)
---------------
22-JUL-2006  MGD  v1.0.0, initial development
*/

/*****************************************************************************/

#define SOFTWAREVN "1.0.0"
#define SOFTWARE_COPYRIGHT "Copyright (C) 2006 Mark G.Daniel - GPL licensed"
#define SOFTWARENM "DYNDNSUPD"
#ifdef __ALPHA
#  define SOFTWAREID SOFTWARENM " AXP-" SOFTWAREVN
#endif
#ifdef __ia64
#  define SOFTWAREID SOFTWARENM " IA64-" SOFTWAREVN
#endif
#ifdef __VAX
#  define SOFTWAREID SOFTWARENM " VAX-" SOFTWAREVN
#endif

#define AGENT_AUTHOR "mark.daniel@wasd.vsm.com.au"

#ifndef __VAX
#   pragma nomember_alignment
#endif

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <unistd.h>

#include <socket.h>
#include <inet.h>
#include <in.h>

#include <descrip.h>
#include <jpidef.h>
#include <libdef.h>
#include <libdtdef.h>
#include <lib$routines.h>
#include <lnmdef.h>
#include <psldef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <starlet.h>

#include <openssl/crypto.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

/* these make it easier to have both VAX and Alpha declarations */
#define OPC$_RQ_RQST 3
#define OPC$M_OPR_CENTRAL 0x1
#define OPC$M_OPR_PRINTER 0x2
#define OPC$M_OPR_TAPES 0x4
#define OPC$M_OPR_DISKS 0x8
#define OPC$M_OPR_DEVICES 0x10
#define OPC$M_OPR_CARDS 0x20
#define OPC$M_OPR_NETWORK 0x40
#define OPC$M_OPR_CLUSTER 0x80
#define OPC$M_OPR_SECURITY 0x100
#define OPC$M_OPR_REPLY 0x200
#define OPC$M_OPR_SOFTWARE 0x400
#define OPC$M_OPR_LICENSE 0x800
#define OPC$M_OPR_USER1 0x1000
#define OPC$M_OPR_USER2 0x2000
#define OPC$M_OPR_USER3 0x4000
#define OPC$M_OPR_USER4 0x8000
#define OPC$M_OPR_USER5 0x10000
#define OPC$M_OPR_USER6 0x20000
#define OPC$M_OPR_USER7 0x40000
#define OPC$M_OPR_USER8 0x80000
#define OPC$M_OPR_USER9 0x100000
#define OPC$M_OPR_USER10 0x200000
#define OPC$M_OPR_USER11 0x400000
#define OPC$M_OPR_USER12 0x800000

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))

#define BOOL int
#define TRUE 1
#define FALSE 0

#define DEFAULT_FORCE        14  /* days */
#define DEFAULT_PERIOD       30  /* minutes */
#define DEFAULT_PRCNAM       "DynDNSupd"

#define DEFAULT_WILDCARD     "NOCHG"
#define DEFAULT_MX           "NOCHG"
#define DEFAULT_BACKMX       "NOCHG"

#define DEFAULT_HOST_CHECK   "checkip.dyndns.org:80"
#define DEFAULT_PORT_CHECK   80
#define DEFAULT_HOST_UPDATE  "members.dyndns.org:443"
#define DEFAULT_PORT_UPDATE  443

#define LOGNAME_AUTH         "DYNDNSUPD_AUTH"
#define LOGNAME_HOST_CHECK   "DYNDNSUPD_HOST_CHECK"
#define LOGNAME_HOST_UPDATE  "DYNDNSUPD_HOST_UPDATE"
#define LOGNAME_HOST         "DYNDNSUPD_HOST"
#define LOGNAME_IP           "DYNDNSUPD_IP"
#define LOGNAME_LOG          "DYNDNSUPD_LOG"
#define LOGNAME_NOW          "DYNDNSUPD_NOW"
#define LOGNAME_PARAM        "DYNDNSUPD_PARAM"
#define LOGNAME_PID          "DYNDNSUPD_PID"
#define LOGNAME_RETURN       "DYNDNSUPD_RETURN"
#define LOGNAME_TIME_LAST    "DYNDNSUPD_TIME_LAST"
#define LOGNAME_TIME_NEXT    "DYNDNSUPD_TIME_NEXT"
#define LOGNAME_TIME_UPDATE  "DYNDNSUPD_TIME_UPDATE"

/******************/
/* global storage */
/******************/

int  Debug,
     DebugWatch,
     ForceUpdate = DEFAULT_FORCE,
     NoLog,
     OpcomTarget,
     PeriodMinutes = DEFAULT_PERIOD,
     UpdateNow;

unsigned short  PortCheck,
                PortUpdate;

unsigned long  CurrentBinTime [2];

char  AuthBase64 [128],
      AuthPlain [128],
      CurrentIpAddrStr [32],
      CurrentTimeStr [24],
      HostCheck [128],
      HostUpdate [128],
      NextTimeStr [24],
      ParamHostName [256],
      ParamWildCard [16],
      ParamMx [96],
      ParamBackMx [16],
      PrcNamStr [16] = DEFAULT_PRCNAM,
      SoftwareID [96];

$DESCRIPTOR (DateTimeFaoDsc, "!20%D\0");

/***********************/
/* DynDNS return codes */
/***********************/

/* look familiar :-) */
#define DYN_RET_SUCCESS 1
#define DYN_RET_WARN    0
#define DYN_RET_ERROR   2
#define DYN_RET_FATAL   4

struct DynDnsCodesStruct {
   int  sev,
        len;
   char *cptr,
        *mptr;
} DynDnsCodes [] = {

   { DYN_RET_ERROR, 6, "badsys",
"the \'system\' parameter specified is not valid" },
   { DYN_RET_FATAL, 8, "badagent",
"this update agent has been blocked for poor behaviour" },
   { DYN_RET_ERROR, 7, "badauth",
"user credentials are incorrect" },
   { DYN_RET_ERROR, 8, "!donator",
"option only available to credited users" },
   { DYN_RET_SUCCESS, 4, "good",
"update was successful" },
   { DYN_RET_WARN, 5, "nochg",
"update changed no settings - more of these will be considered abuse \
and the hostname will be blocked" },
   { DYN_RET_ERROR, 7, "notfqdn",
"the hostname specified is not fully-qualified" },
   { DYN_RET_ERROR, 6, "nohost",
"the hostname specified does not exist" },
   { DYN_RET_ERROR, 6, "!yours",
"the hostname specified does not belong to the username specified" },
   { DYN_RET_ERROR, 7, "numhost",
"too few or too many hosts found" },
   { DYN_RET_FATAL, 5, "abuse",
"the hostname specified has been blocked for update abuse" },
   { DYN_RET_FATAL, 6, "dnserr",
"DNS error encountered" },
   { DYN_RET_FATAL, 3, "911",
"serious DynDNS error" },
   { 0, 0, NULL, NULL }

};

/**************/
/* prototypes */
/**************/

int EncodeBase64 (char*, int, char*, int);
char* CheckIpAddr ();
int DaysSinceUpdate ();
int DefPidLogName ();
int GetParameters ();
int IpAddrHasChanged ();
int ReportOpcom (int, char*, unsigned long*);
void SchedWakeup ();
char* SysTrnLnm (char*, char*);
int SysCreLnm (char*, char*);
int SysDelLnm (char*);
int UpdateRequest ();
char* UpdateSSL (char*);
void UponExit ();
void LogThis (int, char, char*, ...);
int verify_callback(int ok, X509_STORE_CTX *ctx);
int strsame (char*, char*, int);
int strzcpy (char*, char*, int);

/*****************************************************************************/
/*
*/

int main()

{
   static $DESCRIPTOR (CurrentTimeDsc, CurrentTimeStr);
   static $DESCRIPTOR (PrcNamDsc, PrcNamStr);
   static $DESCRIPTOR (PrcNamFaoDsc, "!AZ:!AZ\0");

   int  retval, status;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   sprintf (SoftwareID, "%s", SOFTWAREID);

   PrcNamDsc.dsc$w_length = strlen(PrcNamStr);
   status = sys$setprn (&PrcNamDsc);

   Debug = (SysTrnLnm ("DYNDNSUPD$DBUG", NULL) != NULL);
   DebugWatch = (SysTrnLnm ("DYNDNSUPD$WATCH", NULL) != NULL);

   cptr = SysTrnLnm (LOGNAME_LOG, NULL);
   if (!NoLog && cptr)
   {
      stdout = freopen (cptr, "a", stdout, "ctx=rec","shr=put");
      if (!stdout) exit (vaxc$errno);
   }

   /* basically so we can check if OPCOM is enabled */
   GetParameters ();

   atexit (UponExit);

   LogThis (__LINE__, 'W', "STARTUP !AZ on !11%D", SoftwareID, 0);
   LogThis (__LINE__, 'I', "LICENSE !AZ", SOFTWARE_COPYRIGHT);

   status = DefPidLogName ();
   if (VMSnok (status))
   {
      LogThis (__LINE__, 'F', "DefPidLogName() %X!8XL", status);
      exit (status);
   }

   if (!AuthPlain[0] && !SysTrnLnm (LOGNAME_AUTH, NULL))
   {
      LogThis (__LINE__, 'F', LOGNAME_AUTH "?");
      exit (SS$_ABORT);
   }

   /* (try to) establish the IP address at the last update */
   cptr = SysTrnLnm (LOGNAME_IP, NULL);
   if (cptr) strzcpy (CurrentIpAddrStr, cptr, sizeof(CurrentIpAddrStr));

   for (;;)
   {
      /********/
      /* loop */
      /********/

      sys$gettim (&CurrentBinTime);
      sys$fao (&DateTimeFaoDsc, 0, &CurrentTimeDsc, &CurrentBinTime);

      status = SysCreLnm (LOGNAME_TIME_LAST, CurrentTimeStr);
      if (VMSnok (status))
      {
         LogThis (__LINE__, 'E', "SysCreLnm() %X!8XL", status);
         exit (status);
      }

      /* parameter defaults */
      strcpy (ParamWildCard, DEFAULT_WILDCARD);
      strcpy (ParamMx, DEFAULT_MX);
      strcpy (ParamBackMx, DEFAULT_BACKMX);

      retval = GetParameters ();
      if (!retval)
      {
         LogThis (__LINE__, 'F', "PARAMETERS?");
         exit (SS$_BUGCHECK);
      }

      if (!ParamHostName[0])
      {
         LogThis (__LINE__, 'F', "/HOSTNAME=?");
         exit (SS$_BUGCHECK);
      }

      if (!(strsame(ParamWildCard,"ON",-1) ||
            strsame(ParamWildCard,"OFF",-1) ||
            strsame(ParamWildCard,"NOCHG",-1)))
      {
         LogThis (__LINE__, 'E', "KEYWORD /WILDCARD=\"!AZ\"", ParamWildCard);
         strcpy (ParamWildCard, DEFAULT_WILDCARD);
      }

      if (!(inet_addr(ParamMx) ||
            strsame(ParamMx,"NOCHG",-1)))
      {
         LogThis (__LINE__, 'E', "KEYWORD /MX=\"!AZ\"", ParamMx);
         strcpy (ParamMx, DEFAULT_WILDCARD);
      }

      if (!(strsame(ParamBackMx,"YES",-1) ||
            strsame(ParamBackMx,"NO",-1) ||
            strsame(ParamBackMx,"NOCHG",-1)))
      {
         LogThis (__LINE__, 'E', "KEYWORD /BACKMX=\"!AZ\"", ParamBackMx);
         strcpy (ParamBackMx, DEFAULT_BACKMX);
      }

      status = SysCreLnm (LOGNAME_HOST, ParamHostName);
      if (VMSnok (status))
      {
         LogThis (__LINE__, 'E', "SysCreLnm() %X!8XL", status);
         exit (status);
      }

      cptr = SysTrnLnm (LOGNAME_AUTH, NULL);
      if (cptr)
      {
         strzcpy (AuthPlain, cptr, sizeof(AuthPlain));
         SysDelLnm (LOGNAME_AUTH);
      }

      PortCheck = 0;
      cptr = SysTrnLnm (LOGNAME_HOST_CHECK, NULL);
      if (cptr)
         strzcpy (HostCheck, cptr, sizeof(HostCheck));
      else
         strcpy (HostCheck, DEFAULT_HOST_CHECK);
      for (cptr = HostCheck; *cptr && *cptr != ':'; cptr++);
      if (*cptr)
      {
         *cptr++ = '\0';
         if (isdigit(*cptr)) PortCheck = (short)atoi(cptr);
      }
      if (!PortCheck) PortCheck = DEFAULT_PORT_CHECK;

      PortUpdate = 0;
      cptr = SysTrnLnm (LOGNAME_HOST_UPDATE, NULL);
      if (cptr)
         strzcpy (HostUpdate, cptr, sizeof(HostUpdate));
      else
         strcpy (HostUpdate, DEFAULT_HOST_UPDATE);
      for (cptr = HostUpdate; *cptr && *cptr != ':'; cptr++);
      if (*cptr)
      {
         *cptr++ = '\0';
         if (isdigit(*cptr)) PortUpdate = (short)atoi(cptr);
      }
      if (!PortUpdate) PortUpdate = DEFAULT_PORT_UPDATE;

      cptr = SysTrnLnm (LOGNAME_NOW, NULL);
      if (cptr)
      {
         SysDelLnm (LOGNAME_NOW);
         UpdateNow = TRUE;
      }

      /* time for a forced update? */
      retval = DaysSinceUpdate();
      if (retval >= ForceUpdate)
      {
         LogThis (__LINE__, 'U', "FORCE after !UL days", retval);
         UpdateNow = TRUE;
      }

      /*********/
      /* check */
      /*********/

      LogThis (__LINE__, 'I', "CHECK !AZ",
                 CurrentIpAddrStr[0] ? CurrentIpAddrStr : "(none)");

      retval = IpAddrHasChanged ();

      if (UpdateNow)
      {
         /* from CLI or logical name one-shots */
         if (retval >= 0)
         {
            /* only if there was no error reported by the IP address check */
            retval = 1;
            UpdateNow = FALSE;
            LogThis (__LINE__, 'U', "NOW !AZ", CurrentIpAddrStr);
         }
      }

      if (retval < 0)
      {
         /*********/
         /* error */
         /*********/

         cptr = "ERROR";
      }
      else
      if (retval > 0)
      {
         /**********/
         /* update */
         /**********/

         LogThis (__LINE__, 'U', "UPDATE !AZ", CurrentIpAddrStr);

         status = SysCreLnm (LOGNAME_IP, CurrentIpAddrStr);
         if (VMSnok (status))
            LogThis (__LINE__, 'E', "SysCreLnm() %X!8XL", status);

         retval = UpdateRequest ();

         if (retval)
            cptr = "OK";
         else
         {
            cptr = "FAIL";
            /* ensure it's retried next time through */
            CurrentIpAddrStr[0] = '\0';
         }
      }
      else
      {
         /* retval is zero */
         LogThis (__LINE__, 'I', "NOCHANGE !AZ", CurrentIpAddrStr);
         cptr = "NOCHG";
      }

      PrcNamDsc.dsc$w_length = sizeof(PrcNamStr)-1;
      sys$fao (&PrcNamFaoDsc, 0, &PrcNamDsc, DEFAULT_PRCNAM, cptr);
      PrcNamDsc.dsc$w_length = strlen(PrcNamStr);
      status = sys$setprn (&PrcNamDsc);
      if (VMSnok (status))
      {
         LogThis (__LINE__, 'E', "sys$setprn() %X!8XL", status);
         exit (status);
      }

      /*************/
      /* hibernate */
      /*************/

      SchedWakeup ();

      fsync (fileno(stdout));

      sys$hiber ();
   }

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
Of course this will only work for exit, not STOP/ID=, etc.
*/

void UponExit ()

{
   /*********/
   /* begin */
   /*********/

   LogThis (__LINE__, 'W', "EXIT !AZ on !11%D", SoftwareID, 0);
}

/*****************************************************************************/
/*
Schedule a wakeup after the appropriate period.
*/

void SchedWakeup ()

{
   static unsigned long  Addend = 0,
                         OneMinute = -600000000;
   static $DESCRIPTOR (NextTimeDsc, NextTimeStr);

   int  status;
   unsigned long  DeltaBinTime [2],
                  NextBinTime [2];

   /*********/
   /* begin */
   /*********/

   status = lib$emul (&PeriodMinutes, &OneMinute, &Addend, &DeltaBinTime);
   if (Debug) fprintf (stdout, "lib$emul() %%X%08.08X\n", status);

   status = lib$add_times (&CurrentBinTime, &DeltaBinTime, &NextBinTime);
   if (Debug) fprintf (stdout, "lib$add_times() %%X%08.08X\n", status);

   status = sys$schdwk (0, 0, &NextBinTime, 0);
   if (VMSnok (status))
   {
      LogThis (__LINE__, 'E', "sys$schdwk() %X!8XL", status);
      exit (status);
   }

   sys$fao (&DateTimeFaoDsc, 0, &NextTimeDsc, &NextBinTime);
   status = SysCreLnm (LOGNAME_TIME_NEXT, NextTimeStr);
   if (VMSnok (status))
      LogThis (__LINE__, 'E', "SysCreLnm() %X!8XL", status);
}

/*****************************************************************************/
/*
Return the number of days elapsed since the time stored (as a string) in the
logical name containing the date/time of the last update to DynDNS.
*/

int DaysSinceUpdate ()

{
   static unsigned long  LibDeltaDays = LIB$K_DELTA_DAYS;

   int  status;
   unsigned long  DeltaDays;
   unsigned long  ResultBinTime [2],
                  UpdateBinTime [2];
   char  *cptr;
   char  UpdateTimeStr [24];
   $DESCRIPTOR (UpdateTimeDsc, UpdateTimeStr);

   /*********/
   /* begin */
   /*********/

   cptr = SysTrnLnm (LOGNAME_TIME_UPDATE, NULL);
   if (!cptr) return (0);

   strzcpy (UpdateTimeStr, cptr, sizeof(UpdateTimeStr));
   UpdateTimeDsc.dsc$w_length = strlen(UpdateTimeStr);
   status = lib$convert_date_string (&UpdateTimeDsc, &UpdateBinTime,
                                     0, 0, 0, 0);
   if (VMSnok (status))
   {
      LogThis (__LINE__, 'E', "lib$convert_date_string()() %X!8XL", status);
      return (0);
   }
   status = lib$sub_times (&CurrentBinTime, &UpdateBinTime, &ResultBinTime);
   if (VMSnok (status))
   {
      if (status == LIB$_NEGTIM) return (0);
      LogThis (__LINE__, 'E', "lib$sub_times() %X!8XL", status);
      return (0);
   }
   status = lib$cvt_from_internal_time (&LibDeltaDays, &DeltaDays,
                                        &ResultBinTime);
   if (VMSnok (status))
   {
      LogThis (__LINE__, 'E', "lib$sub_times() %X!8XL", status);
      return (0);
   }

   return (DeltaDays);
}

/*****************************************************************************/
/*
Build an HTTP request suitable for the update client described in
http://www.dyndns.com/developers/specs/ and send it to the DynDNS update
server.  Parse the response and detrermine whether the update has been
successful or not.
*/

BOOL UpdateRequest ()

{
   int  idx, slen, status,
        ErrorCount,
        FatalCount,
        SuccessCount,
        WarningCount;
   char  ch;
   char  *cptr, *lptr;
   char  ubuf [2048],
         ReturnValue [128];

   /*********/
   /* begin */
   /*********/

   if (AuthPlain[0])
   {
      EncodeBase64 (AuthPlain, -1, AuthBase64, sizeof(AuthBase64));
      AuthPlain[0] = '\0';
   }

   sprintf (ubuf,
"GET /nic/update?system=%s&hostname=%s&myip=%s&wildcard=%s&mx=%s&backmx=%s&\
offline=%s HTTP/1.0\r\n\
Host: %s\r\n\
Authorization: Basic %s\r\n\
User-Agent: %s %s\r\n\
\r\n",
            "dyndns",
            ParamHostName,
            CurrentIpAddrStr,
            ParamWildCard,
            ParamMx,
            ParamBackMx,
            "NO",
            HostUpdate,
            AuthBase64,
            SoftwareID, AGENT_AUTHOR);

   lptr = UpdateSSL (ubuf);
   if (DebugWatch) LogThis (__LINE__, 'D', "!AZ", lptr ? lptr : "(null)");
   if (!lptr) return (FALSE);

   ErrorCount = FatalCount = SuccessCount = WarningCount = 0;
   while (*lptr)
   {
      for (idx = 0; DynDnsCodes[idx].len; idx++)
      {
         if (strncmp (lptr, DynDnsCodes[idx].cptr, DynDnsCodes[idx].len))
            continue;
         for (cptr = lptr; *cptr && *cptr != '\n' && *cptr != '\r'; cptr++);
         slen = cptr - lptr;
         switch (DynDnsCodes[idx].sev)
         {
            case DYN_RET_SUCCESS : SuccessCount++; cptr = "SUCCESS";
                                   ch = 'U'; break;
            case DYN_RET_WARN : WarningCount++; cptr = "WARNING";
                                ch = 'W'; break;
            case DYN_RET_ERROR : ErrorCount++; cptr = "ERROR";
                                 ch = 'E'; break;
            case DYN_RET_FATAL : FatalCount++; cptr = "FATAL";
                                 ch = 'E'; break;
            default : cptr = "UNKNOWN"; ch = 'E';
         }
         LogThis (__LINE__, ch, "!AZ \"!#AZ\" !AZ",
                  cptr, slen, lptr, DynDnsCodes[idx].mptr);
      }
      /* find start of (any) next line */
      lptr += slen;
      while (*lptr && (*lptr != '\n' && *lptr != '\r')) lptr++;
      while (*lptr && (*lptr == '\n' || *lptr == '\r')) lptr++;
   }

   status = SysCreLnm (LOGNAME_TIME_UPDATE, CurrentTimeStr);
   if (VMSnok (status))
      LogThis (__LINE__, 'E', "SysCreLnm() %X!8XL", status);

   sprintf (ReturnValue, "SUCCESS:%d WARNING:%d ERROR:%d FATAL:%d",
            SuccessCount, WarningCount, ErrorCount, FatalCount);
   status = SysCreLnm (LOGNAME_RETURN, ReturnValue);
   if (VMSnok (status))
      LogThis (__LINE__, 'E', "SysCreLnm() %X!8XL", status);

   if (!ErrorCount && !FatalCount && !SuccessCount && !WarningCount)
      LogThis (__LINE__, 'E', "RETURN CODE(S) not found");

   if (ErrorCount || FatalCount) return (FALSE);

   return (TRUE);
}

/*****************************************************************************/
/*
Using CheckIpAddr() to get this systems current IP address from the DynDNS Web
server compare it to that (if any) stored in 'CurrentIpAddrStr' and return TRUE
if they are the same or FALSE if not.
*/

int IpAddrHasChanged ()

{
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   cptr = CheckIpAddr ();
   if (!cptr) return (-1);
   if (!strcmp (cptr, CurrentIpAddrStr)) return (0);

   if (DebugWatch)
      LogThis (__LINE__, 'U', "CHANGED !AZ to !AZ", CurrentIpAddrStr, cptr);

   zptr = (sptr = CurrentIpAddrStr) + sizeof(CurrentIpAddrStr)-1;
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   return (1);
}

/*****************************************************************************/
/*
Check the current system's IP address by requesting a page from the DynDNS
check-IP Web service and parsing that from the content.  A pointer to a static
string buffer containing the dotted-decimal address is returned if successful.
Any error returns NULL.
*/

char* CheckIpAddr ()

{
   static char  IpAddrStr [32];

   int  cnt, retval, sock, slen,
        HttpStatus;
   char  *cptr, *sptr, *zptr;
   char  sbuf [256],
         rbuf [1024];
   struct sockaddr_in  saddr;
   struct hostent  *heptr;

   /*********/
   /* begin */
   /*********/

   slen = sprintf (sbuf,
"GET / HTTP/1.0\r\n\
Host: %s\r\n\
User-Agent: %s %s\r\n\
\r\n",
                   HostCheck,
                   SoftwareID, AGENT_AUTHOR);

   memset (&saddr, 0, sizeof(saddr));
   saddr.sin_family = AF_INET;
   saddr.sin_port = htons(PortCheck);

   if (isdigit(HostCheck[0]))
      saddr.sin_addr.s_addr = inet_addr(HostCheck);
   else
   {
      heptr = gethostbyname (HostCheck);
      if (!heptr)
      {
         LogThis (__LINE__, 'E', "gethostbyname() !AZ !UL %X!8XL",
                    HostCheck, errno, vaxc$errno);
         return (NULL);
      }
      memcpy (&saddr.sin_addr.s_addr, heptr->h_addr, 4);
   }

   sock = socket (PF_INET, SOCK_STREAM, IPPROTO_TCP);       
   if (sock == -1)
   {
      LogThis (__LINE__, 'E', "socket() !SL %X!8XL", errno, vaxc$errno);
      return (NULL);
   }

   if (DebugWatch)
      LogThis (__LINE__, 'D', "CONNECT !AZ:!UL", HostCheck, PortCheck);
   retval = connect (sock, (struct sockaddr*)&saddr, sizeof(saddr)); 
   if (retval == -1)
   {
      LogThis (__LINE__, 'E', "CONNECT !AZ:!UL !SL %X!8XL",
                 HostCheck, PortCheck, errno, vaxc$errno);
      return (NULL);
   }

   if (DebugWatch) LogThis (__LINE__, 'D', "REQUEST\r\n!AZ", sbuf);

   retval = write (sock, sbuf, slen);  
   if (retval == -1)
   {
      LogThis (__LINE__, 'E', "write() !SL %X!8XL", errno, vaxc$errno);
      return (NULL);
   }

   sptr = rbuf;
   cnt = sizeof(rbuf)-1;
   while (cnt)
   {
      retval = read (sock, sptr, cnt);
      if (Debug) fprintf (stdout, "read() %d", retval);
      if (retval <= 0) break;
      sptr += retval;
      cnt -= retval;
   }
   *sptr = '\0';
   slen = sptr - rbuf;
   if (DebugWatch) LogThis (__LINE__, 'D', "RECEIVED !UL\r\n!AZ", slen, rbuf);

   retval = close(sock);
   if (retval == -1)
   {
      LogThis (__LINE__, 'E', "close() !SL %X!8XL", errno, vaxc$errno);
      return (NULL);
   }

   /* ensure we've got what looks like an HTTP response and it's 200 */
   cptr = rbuf;
   while (*cptr && !isspace(*cptr)) cptr++;
   while (*cptr && isspace(*cptr)) cptr++;
   HttpStatus = atoi(cptr);
   if (strncmp (rbuf, "HTTP/1.", 7) || HttpStatus != 200)
   {
      while (*cptr && *cptr != '\n' && *cptr != '\n') cptr++;
      *cptr = '\0';
      LogThis (__LINE__, 'E', "RESPONSE !AZ", rbuf);
      return (NULL);
   }

   zptr = (sptr = IpAddrStr) + sizeof(IpAddrStr)-1;
   cptr = strstr (rbuf, "</head><body>");
   if (!cptr) return (NULL);
   while (*cptr && !isdigit(*cptr)) cptr++;
   if (!*cptr) return (NULL);
   while (*cptr && (isdigit(*cptr) || *cptr == '.') && sptr < zptr)
      *sptr++ = *cptr++;
   *sptr = '\0';

   if (inet_addr(HostCheck) == 0) return (NULL);

   if (DebugWatch) LogThis (__LINE__, 'D', "CURRENT !AZ", IpAddrStr);

   return (IpAddrStr);
}

/*****************************************************************************/
/*
This function uses SSL and connects to the DynDNS secure update service.  When
connected it sends the HTTP request priovided via parameter 'rptr'.  If the
request receives a response a pointer to the body of that response is returned. 
If any sort of error occured during request setup or processing a NULL pointer
is returned.
*/

char* UpdateSSL (char *rptr)

{
   static char  ResponseBuffer [256];

   int  cnt, retval, slen, sock,
        HttpStatus;
   char  *cptr, *sptr, *zptr;
   char  rbuf [1024];
   struct sockaddr_in  saddr;
   struct hostent  *heptr;

   SSL_CTX  *ctx;
   SSL  *ssl;
   SSL_METHOD  *meth;
   X509  *ServerCertPtr;

   /*********/
   /* begin */
   /*********/

   memset (&saddr, 0, sizeof(saddr));
   saddr.sin_family = AF_INET;
   saddr.sin_port = htons(PortUpdate);

   if (isdigit(HostUpdate[0]))
      saddr.sin_addr.s_addr = inet_addr(HostUpdate);
   else
   {
      heptr = gethostbyname (HostUpdate);
      if (!heptr)
      {
         LogThis (__LINE__, 'E', "gethostbyname() !AZ !SL %X!8XL",
                    HostUpdate, errno, vaxc$errno);
         return (NULL);
      }
      memcpy (&saddr.sin_addr.s_addr, heptr->h_addr, 4);
   }

   /************/
   /* SSL init */
   /************/

   SSL_library_init();
   SSL_load_error_strings();
   meth = SSLv3_method();
   ctx = SSL_CTX_new(meth);                        
   if (!ctx)
   {
      LogThis (__LINE__, 'E', "SSL_new() !SL %X!8XL", errno, vaxc$errno);
      ERR_print_errors_fp (stdout);
      return (NULL);
   }

   /***********/
   /* connect */
   /***********/

   sock = socket (PF_INET, SOCK_STREAM, IPPROTO_TCP);       
   if (sock == -1)
   {
      LogThis (__LINE__, 'E', "socket() !SL %X!8XL", errno, vaxc$errno);
      return (NULL);
   }

   if (DebugWatch)
      LogThis (__LINE__, 'D', "CONNECT !AZ:!UL", HostUpdate, PortUpdate);
   retval = connect (sock, (struct sockaddr*)&saddr, sizeof(saddr)); 
   if (retval == -1)
   {
      LogThis (__LINE__, 'E', "CONNECT !AZ:!UL !SL %X!8XL",
                 HostUpdate, PortUpdate, errno, vaxc$errno);
      close (sock);
      return (NULL);
   }

   ssl = SSL_new (ctx);
   if (!ssl)
   {
      LogThis (__LINE__, 'E', "SSL_new() !SL %X!8XL", errno, vaxc$errno);
      ERR_print_errors_fp (stdout);
      close (sock);
      return (NULL);
   }

   SSL_set_fd (ssl, sock);

   retval = SSL_connect(ssl);
   if (retval == -1)
   {
      LogThis (__LINE__, 'E', "SSL_connect() !SL %X!8XL", errno, vaxc$errno);
      ERR_print_errors_fp (stdout);
      SSL_free(ssl);
      close (sock);
      return (NULL);
   }

   if (DebugWatch)
   {
      LogThis (__LINE__, 'D', "CIPHER !AZ", SSL_get_cipher (ssl));
      ServerCertPtr = SSL_get_peer_certificate (ssl);       
      if (ServerCertPtr)
      {
         char  *subptr, *issptr;
         subptr = X509_NAME_oneline(X509_get_subject_name(ServerCertPtr),0,0);
         issptr = X509_NAME_oneline(X509_get_issuer_name(ServerCertPtr),0,0);
         if (subptr && issptr)
            LogThis (__LINE__, 'D', "CERTIFICATE\r\n!AZ\r\n!AZ",
                     subptr, issptr);
         if (subptr) free (subptr);
         if (issptr) free (issptr);
         X509_free (ServerCertPtr);
      }
      else
         LogThis (__LINE__, 'D', "CERTIFICATE (none)");
   }

   /***********/
   /* request */
   /***********/

   if (DebugWatch) LogThis (__LINE__, 'D', "REQUEST\r\n!AZ", rptr);

   retval = SSL_write (ssl, rptr, strlen(rptr));  
   if (retval == -1)
   {
      LogThis (__LINE__, 'E', "SSL_write() !SL %X!8XL", errno, vaxc$errno);
      ERR_print_errors_fp (stdout);
      SSL_CTX_free(ctx);
      SSL_free(ssl);
      close (sock);
      return (NULL);
   }

   /************/
   /* response */
   /************/

   sptr = rbuf;
   cnt = sizeof(rbuf)-1;
   while (cnt)
   {
      retval = SSL_read (ssl, sptr, cnt);
      if (DebugWatch) LogThis (__LINE__, 'D', "read() !SL", retval);
      if (retval <= 0) break;
      sptr += retval;
      cnt -= retval;
   }
   *sptr = '\0';
   slen = sptr - rbuf;
   if (DebugWatch) LogThis (__LINE__, 'D', "!SL\r\n!AZ", slen, rbuf);

   /***************/
   /* SSL closure */
   /***************/

   retval = SSL_shutdown(ssl);
   if (retval == -1)
   {
      LogThis (__LINE__, 'E', "SSL_shutdown() !SL %X!8XL", errno, vaxc$errno);
      ERR_print_errors_fp (stdout);
      SSL_CTX_free(ctx);
      SSL_free(ssl);
      close (sock);
      return (NULL);
   }

   retval = close (sock);
   if (retval == -1)
   {
      LogThis (__LINE__, 'E', "close() !SL %X!8XL", errno, vaxc$errno);
      SSL_CTX_free(ctx);
      SSL_free(ssl);
      close (sock);
      return (NULL);
   }

   SSL_free(ssl);
   SSL_CTX_free(ctx);

   /********************/
   /* process response */
   /********************/

   /* ensure we've got what looks like an HTTP response and it's 200 */
   cptr = rbuf;
   while (*cptr && !isspace(*cptr)) cptr++;
   while (*cptr && isspace(*cptr)) cptr++;
   HttpStatus = atoi(cptr);
   if (strncmp (rbuf, "HTTP/1.", 7) || HttpStatus != 200)
   {
      while (*cptr && *cptr != '\n' && *cptr != '\n') cptr++;
      *cptr = '\0';
      LogThis (__LINE__, 'E', "RESPONSE !AZ", rbuf);
      return (NULL);
   }

   for (cptr = rbuf; *cptr; cptr++)
   {
      if (*cptr != '\n' && *cptr != '\r') continue;
      if (*(unsigned short*)cptr == '\n\n') { cptr += 2; break; }
      if (*(unsigned long*)cptr == '\r\n\r\n') { cptr += 4; break; }
   }

   if (!*cptr)
   {
      LogThis (__LINE__, 'E', "EMPTY response");
      return (NULL);
   }

   zptr = (sptr = ResponseBuffer) + sizeof(ResponseBuffer)-1;
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   if (sptr >= zptr)
   {
      LogThis (__LINE__, 'E', "OVERFLOW");
      return (NULL);
   }

   return (ResponseBuffer);
}

/*****************************************************************************/
/*
Using OpenSSL BIO streams encode a plain-string into base-64.
*/

int EncodeBase64
(
char *TextPtr,
int TextLength,
char *BufferPtr,
int SizeOfBuffer
)
{
   static BIO  *bio,
               *BioB64Ptr,
               *BioMemPtr;
   int  rc;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "EnocdeBase64()\n");

   if (!BioMemPtr)
   {
      BioMemPtr = BIO_new (BIO_s_mem());
      if (!BioMemPtr) exit (SS$_BUGCHECK);
      BioB64Ptr = BIO_new (BIO_f_base64());
      if (!BioB64Ptr) exit (SS$_BUGCHECK);
      BIO_set_flags (BioB64Ptr, BIO_FLAGS_BASE64_NO_NL);
      bio = BIO_push (BioB64Ptr, BioMemPtr);
      if (!bio) exit (SS$_BUGCHECK);
   }

   if (TextLength == -1) TextLength = strlen(TextPtr);
   BIO_reset (bio);
   rc = BIO_write (bio, TextPtr, TextLength);
   rc = BIO_flush (bio);
   rc = BIO_pending (bio);
   if (rc > SizeOfBuffer) rc = SizeOfBuffer;
   rc = BIO_read (BioMemPtr, BufferPtr, rc);
   if (Debug) fprintf (stdout, "%d |%s|\n", rc, BufferPtr);

   return (rc);
}  

/*****************************************************************************/
/*
Put and entry into the log file and if enabled send a similar entry to OPCOM. 
A log file entry has the folowing (basic) format:

  <char> <4-digit-line> <date> <time> <keyword> <string>

The <char> is either 'I' for informational, 'W' for warning, 'E' for transient
error, or 'F' for fatal error.  The <4-digit-line> is the source code line
number making the entry (makes it easier for bug analysis).  The <date> and
<time> of the entry in VMS format.  The <keyword> is hopefully an obvious
indicator or the entry.  The <string> is some free-form text providing
additional information.

An OPCOM message looks like:

  %%%%%%%%%%%  OPCOM   6-AUG-2006 00:00:07.29  %%%%%%%%%%%
  Message from user SYSTEM on KLAATU
  Process DynDNSupd:NOCHG [524] reports
  CHECK 59.167.161.58
*/ 

void LogThis
(
int SourceCodeLine,
char CategoryChar,
char *FormatString,
...
)
{
   static $DESCRIPTOR (LineNumFaoDsc, "!ZL\0");

   int  argcnt, status;
   unsigned short  ShortLength;
   unsigned long  *vecptr;
   unsigned long  BinTime [2];
   unsigned long  FaoVector [31+1];
   char  *cptr, *sptr, *zptr;
   char  CategoryStr [2],
         FormatBuffer [256],
         LineNum [16],
         WatchBuffer [1024];
   $DESCRIPTOR (FaoDsc, "");
   $DESCRIPTOR (LineNumDsc, LineNum);
   $DESCRIPTOR (WatchBufferDsc, WatchBuffer);
   va_list  argptr;

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug) fprintf (stdout, "LogThis() %d\n", argcnt);

   CategoryStr[0] = CategoryChar;
   CategoryStr[1] = '\0';

   zptr = (sptr = FormatBuffer) + sizeof(FormatBuffer)-1;
   for (cptr = "!AZ !4ZL !%D "; *cptr && sptr < zptr; *sptr++ = *cptr++);
   for (cptr = FormatString; *cptr && sptr < zptr; *sptr++ = *cptr++);
   if (sptr < zptr) *sptr++ = '\n';
   *sptr = '\0';
   FaoDsc.dsc$w_length = sptr - FormatBuffer;

   vecptr = FaoVector;
   *vecptr++ = (unsigned long)&CategoryStr;
   *vecptr++ = SourceCodeLine;
   *vecptr++ = 0;
   va_start (argptr, FormatString);
   for (argcnt -= 3; argcnt; argcnt--)
      *vecptr++ = va_arg (argptr, unsigned long);
   va_end (argptr);

   FaoDsc.dsc$a_pointer = FormatBuffer;
   FaoDsc.dsc$w_length = strlen(FormatBuffer);

   status = sys$faol (&FaoDsc, &ShortLength, &WatchBufferDsc, &FaoVector);
   if (VMSnok (status)) exit (status);

   WatchBuffer[ShortLength] = '\0';
   fputs (WatchBuffer, stdout);

   if (OpcomTarget)
   {
      sys$fao (&LineNumFaoDsc, 0, &LineNumDsc, SourceCodeLine);
      zptr = (sptr = FormatBuffer) + sizeof(FormatBuffer)-1;
      for (cptr = "Process "; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = PrcNamStr; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = " ["; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = LineNum; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = "] reports\r\n"; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = FormatString; *cptr && sptr < zptr; *sptr++ = *cptr++);
      *sptr = '\0';

      ReportOpcom (OpcomTarget, FormatBuffer, &FaoVector[3]);
   }
}

/****************************************************************************/
/*
$FAO formatted print statement to OPCOM.  A fixed-size, internal buffer of 986
bytes maximum is used and the result output as an OPCOM message.
*/

int ReportOpcom
(
int OpcomTarget,
char *FormatString,
unsigned long *FaoVector
)
{
   static $DESCRIPTOR (FaoDsc, "");
   static $DESCRIPTOR (OpcomDsc, "");
   static $DESCRIPTOR (OpcomMsgDsc, "");      

   int  status,
        argcnt;
   unsigned short  ShortLength;
   struct
   {
      unsigned long  TargetType;
      unsigned long  RequestId;
      char  MsgText [986+1];
   } OpcomMsg;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ReportOpcom() |%s|\n", FormatString);

   FaoDsc.dsc$a_pointer = FormatString;
   FaoDsc.dsc$w_length = strlen(FormatString);

   OpcomMsgDsc.dsc$a_pointer = (char*)&OpcomMsg.MsgText;
   OpcomMsgDsc.dsc$w_length = sizeof(OpcomMsg.MsgText)-1;

   status = sys$faol (&FaoDsc, &ShortLength, &OpcomMsgDsc, FaoVector);
   if (VMSnok (status)) exit (status);

   OpcomMsg.MsgText[ShortLength] = '\0';
   if (Debug) fprintf (stdout, "%d |%s|\n", ShortLength, OpcomMsg.MsgText);

   OpcomMsg.TargetType = OPC$_RQ_RQST + ((OpcomTarget & 0xffffff) << 8);
   OpcomMsg.RequestId = 0;

   OpcomDsc.dsc$a_pointer = (char*)&OpcomMsg;
   OpcomDsc.dsc$w_length = ShortLength + 8;

   status = sys$sndopr (&OpcomDsc, 0);
   if (VMSnok (status)) exit (status);

   return (status);
}

/*****************************************************************************/
/*
Translate an exec mode logical name using LNM$FILE_DEV.  Return a pointer to
the value string, or NULL if the name does not exist.  If 'LogValue' is
supplied the logical name is translated into that (assumed to be large enough),
otherwise it's translated into an internal static buffer.
*/

char* SysTrnLnm
(
char *LogName,
char *LogValue
)
{
   static unsigned char  ExecMode = PSL$C_EXEC;
   static unsigned short  ShortLength;
   static char  StaticLogValue [256];
   static $DESCRIPTOR (LogNameDsc, "");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");
   static struct {
      short int  buf_len;
      short int  item;
      void  *buf_addr;
      unsigned short  *ret_len;
   } LnmItems [] =
   {
      { 255, LNM$_STRING, 0, &ShortLength },
      { 0,0,0,0 }
   };

   int  status;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SysTrnLnm() |%s|\n", LogName);

   LogNameDsc.dsc$a_pointer = LogName;
   LogNameDsc.dsc$w_length = strlen(LogName);
   if (LogValue)
      cptr = LnmItems[0].buf_addr = LogValue;
   else
      cptr = LnmItems[0].buf_addr = StaticLogValue;

   status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, &ExecMode, &LnmItems);
   if (Debug) fprintf (stdout, "sys$trnlnm() %%X%08.08X\n", status);
   if (!(status & 1))
   {
      if (Debug) fprintf (stdout, "|(null)|\n");
      return (NULL);
   }

   cptr[ShortLength] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", cptr);
   return (cptr);
}

/*****************************************************************************/
/*
Create an exec mode logical name in the system table.  If that name contains
<LF>-delimtted lines the create a translation index for each (0..127).
*/

int SysCreLnm
(
char *LogName,
char *LogValue
)
{
#define LNM_INDEX_MAX 127

   static unsigned char  ExecMode = PSL$C_EXEC;
   static $DESCRIPTOR (LogTableDsc, "LNM$SYSTEM");
   static $DESCRIPTOR (LogNameDsc, "");

   int  idx, status;
   char  *cptr, *sptr;
   struct {
      short int  buf_len;
      short int  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   }
   CreLnmItem [LNM_INDEX_MAX+1];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SysCreLnm() |%s|%s|\n", LogName, LogValue);

   idx = 0;
   cptr = LogValue;
   while (*cptr && idx < LNM_INDEX_MAX)
   {
      for (sptr = cptr; *sptr && *sptr != '\n'; sptr++);
      CreLnmItem[idx].item = LNM$_STRING;
      CreLnmItem[idx].buf_addr = (unsigned char*)cptr;
      CreLnmItem[idx++].buf_len = sptr - cptr;
      if (Debug) fprintf (stdout, "|%.*s|\n", sptr-cptr, cptr); 
      cptr = sptr;
      if (*(unsigned short*)cptr == '\r\n') cptr += 2;
   }
   memset (&CreLnmItem[idx], 0, sizeof(CreLnmItem[idx]));
   LogNameDsc.dsc$a_pointer = LogName;
   LogNameDsc.dsc$w_length = strlen(LogName);
   status = sys$crelnm (0, &LogTableDsc, &LogNameDsc, &ExecMode, &CreLnmItem);
   if (Debug) fprintf (stdout, "sys$crelnm() %%X%08.08X\n", status);
   return (status);
}

/*****************************************************************************/
/*
Delete an exec mode logical name in the system table.
*/

int SysDelLnm (char *LogName)

{
   static unsigned char  ExecMode = PSL$C_EXEC;
   static $DESCRIPTOR (LogTableDsc, "LNM$SYSTEM");
   static $DESCRIPTOR (LogNameDsc, "");

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SysDelLnm() |%s|\n", LogName);

   LogNameDsc.dsc$a_pointer = LogName;
   LogNameDsc.dsc$w_length = strlen(LogName);
   status = sys$dellnm (&LogTableDsc, &LogNameDsc, &ExecMode);
   if (Debug) fprintf (stdout, "sys$dellnm() %%X%08.08X\n", status);

   return (status);
}

/*****************************************************************************/
/*
*/

int DefPidLogName ()

{
   static unsigned long  JpiPid;

   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char  *buf_addr;
      unsigned short  *short_ret_len;
   }
      JpiItems [] =
   {
      { sizeof(JpiPid), JPI$_PID, (void*)&JpiPid, 0 },
      { 0,0,0,0 }
   };

   int  status;
   unsigned long  ProcessContext = 0;
   char  PidStr [16];
   struct {
      unsigned short  iosb$w_status;
      unsigned short  iosb$w_bcnt;
      unsigned long  iosb$l_reserved;
   } IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DefPidLogName()\n");

   status = sys$getjpiw (0, &ProcessContext, 0, &JpiItems, &IOsb, 0, 0);
   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status)) return (status);

   sprintf (PidStr, "%08.08X", JpiPid);
   status = SysCreLnm (LOGNAME_PID, PidStr);
   return (status);
}

/*****************************************************************************/
/*
Get "command-line" parameters, whether from the command-line of the translated
logical name value.  If taken from a logical name value this function rebuilds
the value sans any authorization credentials and recreates the logical name
value.
*/

int GetParameters ()

{
   static unsigned long  Flags = 0;

   BOOL  FromCli;
   int  status,
        ParamCount;
   unsigned short  Length;
   char  ch;
   char  *aptr, *cptr, *clptr, *sptr, *zptr;
   char  CommandLine [256];
   $DESCRIPTOR (CommandLineDsc, CommandLine);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetParameters()\n");

   if (SysTrnLnm (LOGNAME_PARAM, CommandLine))
      FromCli = FALSE;
   else
   {
      /* get the entire command line following the verb */
      status = lib$get_foreign (&CommandLineDsc, 0, &Length, &Flags);
      if (VMSnok (status)) exit (status);

      CommandLine[Length] = '\0';
      FromCli = TRUE;
   }

   LogThis (__LINE__, 'I', "PARAMETERS !AZ", CommandLine);

   ParamCount = 0;
   aptr = NULL;
   clptr = CommandLine;
   ch = *clptr;
   for (;;)
   {
      if (aptr && *aptr == '/') *aptr = '\0';
      if (!ch) break;

      *clptr = ch;
      if (Debug) fprintf (stdout, "clptr |%s|\n", clptr);
      while (*clptr && isspace(*clptr)) *clptr++ = '\0';
      aptr = clptr;
      if (*clptr == '/') clptr++;
      while (*clptr && !isspace (*clptr) && *clptr != '/')
      {
         if (*clptr != '\"')
         {
            clptr++;
            continue;
         }
         cptr = clptr;
         clptr++;
         while (*clptr)
         {
            if (*clptr == '\"')
               if (*(clptr+1) == '\"')
                  clptr++;
               else
                  break;
            *cptr++ = *clptr++;
         }
         *cptr = '\0';
         if (*clptr) clptr++;
      }
      ch = *clptr;
      if (*clptr) *clptr = '\0';
      if (Debug) fprintf (stdout, "aptr |%s|\n", aptr);
      if (!*aptr) continue;

      ParamCount++;

      if (strsame (aptr, "/AUTH=", 5))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         cptr++;
         strzcpy (AuthPlain, cptr, sizeof(AuthPlain));
         continue;
      }

      if (strsame (aptr, "/BACKMX=", 4))
      {
         strcpy (ParamBackMx, "YES");
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         cptr++;
         strzcpy (ParamBackMx, cptr, sizeof(ParamBackMx));
         continue;
      }

      if (strsame (aptr, "/DBUG", -1))
      {
         Debug = TRUE;
         continue;
      }

      if (strsame (aptr, "/FORCE", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         cptr++;
         ForceUpdate = atoi(cptr);
         if (ForceUpdate <= 1) ForceUpdate = DEFAULT_FORCE;
         continue;
      }

      if (strsame (aptr, "/HOSTNAME=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         cptr++;
         /* ensure it contains only alpha-numerics and periods */
         zptr = (sptr = ParamHostName) + sizeof(ParamHostName)-1;
         while (*cptr && sptr < zptr)
         {
            if (isalnum(*cptr) || *cptr == '.'  || *cptr == '-' || *cptr == ',')
               *sptr++ = *cptr++;
            else
               cptr++;
         }
         *sptr = '\0';
         continue;
      }

      if (strsame (aptr, "/MX=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         cptr++;
         strzcpy (ParamMx, cptr, sizeof(ParamMx));
         continue;
      }

      if (FromCli && strsame (aptr, "/NOLOG", -1))
      {
         NoLog = TRUE;
         continue;
      }

      if (strsame (aptr, "/NOW", 4))
      {
         UpdateNow = TRUE;
         continue;
      }

      if (strsame (aptr, "/OPCOM=", 4))
      {
         OpcomTarget = OPC$M_OPR_CENTRAL;
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         cptr++;
         OpcomTarget = atoi(cptr);
         continue;
      }

      if (strsame (aptr, "/PERIOD=", 4))
      {
         int  hours = 0, mins = 0;
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         mins = atoi(cptr);
         while (isdigit(*cptr)) cptr++;
         if (*cptr == ':')
         {
            cptr++;
            hours = mins;
            mins = atoi(cptr);
            if (mins > 59) mins = 0;
            PeriodMinutes = hours * 60 + mins;
         }
         else
            PeriodMinutes = mins;
         continue;
      }

      if (strsame (aptr, "/WATCH", -1))
      {
         DebugWatch = TRUE;
         continue;
      }

      if (strsame (aptr, "/WILDCARD=", 4))
      {
         strcpy (ParamWildCard, DEFAULT_WILDCARD);
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         cptr++;
         strzcpy (ParamWildCard, cptr, sizeof(ParamWildCard));
         continue;
      }

      if (FromCli)
      {
         if (*aptr == '/')
         {
            fprintf (stdout,
"%%DYNDNSUPD-E-IVQUAL, unrecognized qualifier\n \\%s\\\n",
                     aptr+1);
            exit (STS$K_ERROR | STS$M_INHIB_MSG);
         }
         else
         {
            fprintf (stdout,
"%%DYNDNSUPD-E-MAXPARM, too many parameters\n \\%s\\\n",
                     aptr);
            exit (STS$K_ERROR | STS$M_INHIB_MSG);
         }
      }
   }

   return (ParamCount);
}

/****************************************************************************/
/*
Copy a string without overflow (or indication of it :-).
*/ 

int strzcpy
(
char *String,
char *cptr,
int SizeOfString
)
{
   char  *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (SizeOfString) SizeOfString--;
   zptr = (sptr = String) + SizeOfString;
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';
   return (sptr - String);
}

/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
TRUE if two strings are the same, or FALSE if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 

BOOL strsame
(
char *sptr1,
char *sptr2,
int  count
)
{
   /*********/
   /* begin */
   /*********/

   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (FALSE);
      if (count)
         if (!--count) return (TRUE);
   }
   if (*sptr1 || *sptr2)
      return (FALSE);
   else
      return (TRUE);
}

/*****************************************************************************/

